-------------------------------------------------------------------------------
--|> Events
-- Manages all events
-------------------------------------------------------------------------------
--|> UPVALUE GLOBALS
-------------------------------------------------------------------------------
local weizPVP = weizPVP
local UnitIsPVPSanctuary = UnitIsPVPSanctuary
local UnitIsEnemy = UnitIsEnemy
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local GetTime = GetTime
local UnitIsPlayer = UnitIsPlayer
local UnitLevel = UnitLevel
local IsInInstance = IsInInstance
local bit_band = bit.band
local C_Timer_After = C_Timer.After
local InCombatLockdown = InCombatLockdown
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo
local _

-------------------------------------------------------------------------------
--|> LIBS
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--|> FUNCTIONS
-------------------------------------------------------------------------------

--=> Static Role Assignment
---------------------------------------------------------------------
local role
function weizPVP:RoleAssign(class)
	role = nil
	if class == "ROGUE" or class == "MAGE" or class == "WARLOCK" or class == "HUNTER" then
		role = "DAMAGER"
	end
	return role
end

--=> Get Spec from Spell
-----------------------------------------------------------------------
local specializationID
function weizPVP:GetSpecFromSpell(SpellID)
	specializationID = weizPVP.SpecDB[SpellID]
	return specializationID
end

--=> Get Role from Spec
-----------------------------------------------------------------------
function weizPVP:GetSpecRole(specid)
	_, _, _, _, role, _, _ = GetSpecializationInfoByID(specid)
	return role
end

--=> Check for Valid Instance
-----------------------------------------------------------------------
local inInstance
local WarModeEnabled
function weizPVP:CheckValidInstance()
	--> Checking for Valid BG
	inInstance = select(2, IsInInstance()) --> Get instance info

	if inInstance == "pvp" and weizPVP.Options.EnabledInBattlegrounds == true then
		return true
	end

	if inInstance == "arena" == true and weizPVP.Options.EnabledInArena == true then
		return true
	end
	return false
end

--=> Check for Valid World Zone
-----------------------------------------------------------------------
function weizPVP:CheckValidWorldZone()
	--> Checking for Valid WPVP Zone
	inInstance = select(2, IsInInstance())
	WarModeEnabled = UnitIsWarModeDesired("player")
	if inInstance == "none" and WarModeEnabled == true then
		return true
	else
		return false
	end
end

--=> Get PVP Zone
-----------------------------------------------------------------------
local ZoneTimerStartTime = GetTime()
local ValidBG
local ValidWorldZone
function weizPVP.GetPVPZone()
	if ZoneTimerStartTime + 1 <= GetTime() then
		ZoneTimerStartTime = GetTime()
		if GetZoneText() == "" then
			C_Timer_After(0.5, weizPVP.GetPVPZone)
			return
		end
		ValidWorldZone = weizPVP:CheckValidWorldZone()
		ValidBG = weizPVP.CheckValidInstance()
		if not InCombatLockdown() then
			if ValidWorldZone or ValidBG then
				weizPVP.OnEnable()
			else
				weizPVP.OnDisable()
			end
		end
	else
		return
	end
end

-------------------------------------------------------------------------------
--|> EVENTS
-------------------------------------------------------------------------------

--=> EVENT: Entering Battleground Instance
-----------------------------------------------------------------------
function weizPVP:PlayerEnteringBattlegroundEvent()
	C_Timer_After(1, weizPVP.GetPVPZone)
	wipe(weizPVP.CurrentNameplates)
end

--=> EVENT: ZoneChangedNewAreaEvent
-----------------------------------------------------------------------
function weizPVP:ZoneChangedNewAreaEvent()
	C_Timer_After(1, weizPVP.GetPVPZone)
end
--=> EVENT: Player Entering World
-----------------------------------------------------------------------
function weizPVP:PlayerEnteringWorldEvent()
	wipe(weizPVP.CurrentNameplates)
	C_Timer_After(1, weizPVP.GetPVPZone)
end

--=> EVENT TIMER; ONE SECOND REPEATING
-----------------------------------------------------------------------
local manageListCounter = 0
local LagModeCounter = 0
local LagMode = false
function weizPVP:OneSecondTimer()
	--> FPS Update
	weizPVP.Display.CurrentFPS = GetFramerate()

	--> Manage list timeouts
	if manageListCounter == 0 then
		weizPVP.ManageListTimeouts()
		manageListCounter = 1
	else
		manageListCounter = 0
	end
	if weizPVP.Options.LagMode.Enabled == true then
		if LagModeCounter >= 4 then
			LagModeCounter = 0
			if weizPVP.Display.CurrentFPS < weizPVP.Options.LagMode.FPSThreshold and weizPVP.NearbyListSize > 12 then
				if LagMode == false then
					--> LAGMODE ON
					weizPVP:UnregisterEvent("NAME_PLATE_UNIT_ADDED", "NameplateAdded")
					weizPVP:UnregisterEvent("NAME_PLATE_UNIT_REMOVED", "NameplateRemoved")
					weizPVP.CurrentNameplates = {}
					LagMode = true
				end
			elseif weizPVP.Display.CurrentFPS > weizPVP.Options.LagMode.FPSThreshold + 4 then
				if LagMode == true then
					--> LAGMODE OFF
					weizPVP:RegisterEvent("NAME_PLATE_UNIT_ADDED", "NameplateAdded")
					weizPVP:RegisterEvent("NAME_PLATE_UNIT_REMOVED", "NameplateRemoved")
					LagMode = false
				end
			end
		else
			LagModeCounter = LagModeCounter + 1
		end
	end
end

--=> EVENT: COMBAT_LOG_UNFILTERED
-----------------------------------------------------------------------
local newDataFound
local playerData
local stealth
local dead
local health
local sanctuary
local timestamp
local specID
local srcKnown = false
local class
local validSource = false
local classColor = {1, 1, 1}
local _, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, arg12, arg13
function weizPVP:CombatLogEvent()
	_, --timestamp
		event,
		_, --hideCaster
		srcGUID,
		srcName,
		srcFlags,
		_, --sourceRaidFlags
		dstGUID,
		dstName,
		dstFlags,
		_, --destRaidFlags
		arg12, --spellID
		arg13, --spellName
		_,
		_,
		_ = CombatLogGetCurrentEventInfo()

	-- if ViragDevTool_AddData then
	-- 	ViragDevTool_AddData(
	-- 		{
	-- 			["arg12"] = arg12,
	-- 			["arg13"] = arg13,
	-- 			["event"] = event
	-- 		},
	-- 		"CLOG: "  .. " - " .. event
	-- 	)
	-- end
	validSource = false
	--> Check for valid source
	if (bit_band(srcFlags, 0x00000040) ~= 0 or weizPVP.Options.GLOBAL_DEBUG) and bit_band(srcFlags, 0x00000400) ~= 0 and srcGUID and srcName then
		validSource = true
	end
	--> Src == Dst
	if validSource and (dstName == srcName) then
		specID = weizPVP:GetSpecFromSpell(arg12)
			if specID ~= nil then
				role = weizPVP:GetSpecRole(specID) --> "DAMAGER", "TANK", or "HEALER"
				if role then
					class = select(2, GetPlayerInfoByGUID(srcGUID))
					weizPVP:UpdatePlayerDB(srcName, class, nil, nil, nil, true, false, role)
				end
			end

		if not weizPVP.PlayerDB[srcName] then
			class = select(2, GetPlayerInfoByGUID(srcGUID))
			weizPVP:UpdatePlayerDB(srcName, class, nil, nil, nil, true, false)
		end
		if not weizPVP.PlayerActiveCache[srcGUID] then
			weizPVP:AddPlayerDataToNearby(srcGUID, false)
		end
		weizPVP:UpdatePlayerActiveCache(srcName, false, nil, false, false, nil, srcGUID)
		return
	end

	--> Update time only for non-estimated players in current list
	srcKnown = false
	timestamp = GetTime()

	-- 115834 --> Shroud of Concealment
	-- 114018 --> Shroud of Concealment (2)
	-- 5215 --> Prowl
	-- 156136 --> Stealthman 52 (Stealth Field)
	-- 66 --> Invisability
	-- 1784 --> Stealth
	-- 11327 --> Vanish

	--> Stealth detect
	if validSource then
		if event == "SPELL_AURA_APPLIED" then
			if (arg12 == 1784 or arg12 == 115834 or arg == 114018 or arg12 == 11327 or arg12 == 5215 or arg12 == 66 or arg12 == 156136) then
				if weizPVP.Options.StealthAlert.Enabled == true then
					sanctuary = UnitIsPVPSanctuary("player")
					if not sanctuary or weizPVP.Options.GLOBAL_DEBUG then --> Show alert only if not in sanctuary for Stealth players
						srcKnown = true
						class = select(2, GetPlayerInfoByGUID(srcGUID))
						classColor = RAID_CLASS_COLORS[class]
						if not classColor then
							classColor = {r = 1, g = 1, b = 1}
						end
						if not weizPVP.PlayerDB[srcName] then
							weizPVP:UpdatePlayerDB(srcName, nil, nil, nil, nil, true, false)
						end
						if not weizPVP.PlayerActiveCache[srcGUID] then
							weizPVP:AddPlayerDataToNearby(srcGUID, false)
						end
						-- weizPVP:UpdatePlayerActiveCache(srcName, newDataFound, nil, true, false, nil, srcGUID)
						weizPVP:UpdatePlayerActiveCache(srcName, false, nil, true, false, nil, srcGUID)
						weizPVP:ShowStealthAlert(arg13 .. " Detected!", srcName, classColor.r, classColor.g, classColor.b)
					end
				end
				return
			end
		end
	end

	--> src known?
	if weizPVP.PlayerActiveCache[srcGUID] and weizPVP.PlayerActiveCache[srcGUID].Estimated == false and event ~= "UNIT_DIED" then
		weizPVP:AddPlayerDataToNearby(srcGUID, false)
		srcKnown = true
		--> dst known?
		if weizPVP.PlayerActiveCache[dstGUID] and weizPVP.PlayerActiveCache[dstGUID].Estimated == false then
			weizPVP:AddPlayerDataToNearby(dstGUID, false)
			if event == "UNIT_DIED" then
				weizPVP:UpdatePlayerActiveCache(dstName, nil, 0, false, true, nil, dstGUID)
			end
			return --> return if both src and dst are known
		end
	end

	--> Source Unit Info
	-------------------------------------------------------------------------------------------------------------------
	if not srcKnown then
		if validSource == true then
			if not weizPVP.PlayerActiveCache[srcGUID] or weizPVP.PlayerActiveCache[srcGUID].Estimated == true then
				newDataFound = false
				role = nil
				stealth = nil
				dead = false

				--> If Player doesn't exist in DB
				if not weizPVP.PlayerDB[srcName] then
					newDataFound, role = weizPVP:GetLevelData(true, event, srcName, srcGUID, arg12, arg13)
				elseif weizPVP.PlayerDB[srcName] and weizPVP.PlayerDB[srcName].Estimated == true then
					--> Player exists, but data is estimated; try to get level based on ability used
					newDataFound, role = weizPVP:GetLevelData(true, event, srcName, srcGUID, arg12, arg13)
				end

				--> Check for spec and role based off spellID
				specID = weizPVP:GetSpecFromSpell(arg12)
				if specID ~= nil then
					role = weizPVP:GetSpecRole(specID) --> "DAMAGER", "TANK", or "HEALER"
					if role then
						weizPVP:UpdatePlayerDB(srcName, nil, nil, nil, nil, true, false, role)
					end
				end

				--> No new data was added or updated, send the few things we do know can update(stealth/dead)
				if newDataFound ~= true then
					weizPVP:UpdatePlayerDB(srcName, nil, nil, nil, nil, true, false)
				--> weizPVP:UpdatePlayerDB(name, class, level, race, guild, estimated, dataUpdated)
				end

				--> UPDATE
				weizPVP:UpdatePlayerActiveCache(srcName, newDataFound, nil, stealth, dead, role, srcGUID)
			elseif weizPVP.PlayerActiveCache[srcGUID] ~= nil then
				--> Check for spec and role based off spellID
				specID = weizPVP:GetSpecFromSpell(arg12)
				if specID ~= nil then
					role = weizPVP:GetSpecRole(specID) --> "DAMAGER", "TANK", or "HEALER"
					if role then
						weizPVP:UpdatePlayerDB(srcName, nil, nil, nil, nil, true, false, role)
					end
				end

				if weizPVP.ActiveList[srcGUID] then
					weizPVP.ActiveList[srcGUID].TimeUpdated = timestamp
				else
					weizPVP:UpdatePlayerActiveCache(srcName, nil, nil, nil, nil, nil, srcGUID)
				end
			end
		end
	end

	--> Destination Unit
	-------------------------------------------------------------------------------------------------------------------
	if (bit_band(dstFlags, 0x00000040) ~= 0 or weizPVP.Options.GLOBAL_DEBUG) and bit_band(dstFlags, 0x00000400) ~= 0 and dstGUID and dstName then
		if not weizPVP.PlayerActiveCache[dstGUID] or weizPVP.PlayerActiveCache[dstGUID].Estimated == true then
			newDataFound = false
			playerData = weizPVP.PlayerDB[dstName]
			stealth = nil
			dead = false
			health = nil
			if not playerData or playerData.Estimated then
				newDataFound = weizPVP:GetLevelData(false, event, dstName, dstGUID, arg12, arg13)
			end
			weizPVP:UpdatePlayerDB(dstName, nil, nil, nil, nil, nil, newDataFound)
			--> weizPVP:UpdatePlayerDB(name, class, level, race, guild, estimated, dataUpdated)
			weizPVP:UpdatePlayerActiveCache(dstName, newDataFound, health, stealth, dead, nil, dstGUID)
		elseif weizPVP.PlayerActiveCache[dstGUID] ~= nil then
			if weizPVP.ActiveList[dstGUID] then
				weizPVP.ActiveList[dstGUID].TimeUpdated = timestamp + 0.001
			else
				weizPVP:UpdatePlayerActiveCache(dstName, nil, nil, nil, nil, nil, dstGUID)
			end
		end
		--> if unit died
		if event == "UNIT_DIED" then
			weizPVP:UpdatePlayerActiveCache(dstName, nil, 0, false, true, nil, dstGUID)
		end
	end
end

--=> EVENT: Player Mouseover
-----------------------------------------------------------------------
function weizPVP:PlayerMouseoverEvent()
	weizPVP:GetUnitData("mouseover")
end

--=> EVENT: Player Target Changed
-- -----------------------------------------------------------------------
function weizPVP:PlayerTargetEvent()
	weizPVP:GetUnitData("target")
	weizPVP:ChangeTargetIcon()
end

--=> EVENT: NAMEPLATE_ADDED
-- -----------------------------------------------------------------------
local GUID
function weizPVP:NameplateAdded(_, unit)
	if (UnitIsEnemy("player", unit) or weizPVP.Options.GLOBAL_DEBUG) and UnitIsPlayer(unit) then
		GUID = UnitGUID(unit)
		--> Update time only
		if weizPVP.PlayerActiveCache[GUID] and weizPVP.PlayerActiveCache[GUID].Estimated == false then
			weizPVP.PlayerActiveCache[GUID].Stealth = false
			weizPVP:AddPlayerDataToNearby(GUID, false)
		else
			weizPVP:GetUnitData(unit)
		end
		weizPVP.CurrentNameplates[GUID] = true
	end
end

--=> EVENT: NAMEPLATE REMOVED
-- -----------------------------------------------------------------------
function weizPVP:NameplateRemoved(_, unit)
	GUID = UnitGUID(unit)
	if weizPVP.PlayerActiveCache[GUID] then
		weizPVP:AddPlayerDataToNearby(GUID, false)
	end
	weizPVP.CurrentNameplates[GUID] = nil
end

--=> EVENT: ENTERING COMBAT
-----------------------------------------------------------------------
function weizPVP:EnteringCombat()
	weizPVP:SetStatusBarMessage("|cFFF4564D +++ IN COMBAT +++|r", false, 0)
end

--=> EVENT: LEAVING COMBAT
-----------------------------------------------------------------------
function weizPVP:LeavingCombat()
	weizPVP:SetStatusBarMessage("|cFF53F493 --- OUT OF COMBAT ---|r", true, 3)
	weizPVP:ManageListTimeouts()
end

--=> EVENT: Player Leveled Up
-----------------------------------------------------------------------
function weizPVP:PlayerLevelUp()
	weizPVP.Player.Level = UnitLevel("player") --> Update player level
end
--=> EVENT: UNIT_HEALTH
-----------------------------------------------------------------------
local UnitGUID = UnitGUID
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local newhealth
local lastUpdated = {}
local UPDATE_INT = 0.4

local function timestampUpdated(guid, newTime)
	if lastUpdated[guid] then
		lastUpdated[guid] = newTime
		return
	end
	lastUpdated[guid] = {}
	lastUpdated[guid] = newTime
end

function weizPVP:UnitHealthEvent(_, unit)
	if unit ~= nil then
		GUID = UnitGUID(unit)
		if GUID and weizPVP.PlayersOnBars[GUID] ~= nil then
			timestamp = GetTime()
			newhealth = UnitHealth(unit) / UnitHealthMax(unit) * 100 or nil
			dead = UnitIsDeadOrGhost(unit) or false
			if newhealth and weizPVP.PlayerActiveCache[GUID] and weizPVP.PlayerActiveCache[GUID].Health ~= newhealth then
				weizPVP.PlayerActiveCache[GUID].Health = newhealth
			end
			if weizPVP.PlayerActiveCache[GUID] and dead ~= weizPVP.PlayerActiveCache[GUID].Dead then
				--> Always update when dead
				weizPVP.PlayerActiveCache[GUID].Health = newhealth
				weizPVP.PlayerActiveCache[GUID].Dead = dead
				weizPVP:AddPlayerDataToNearby(GUID, false)
				timestampUpdated(GUID, timestamp)
				return
			end
			--> Update Throttle
			if not lastUpdated[GUID] or (lastUpdated[GUID] and lastUpdated[GUID] + UPDATE_INT < timestamp) then
				weizPVP:AddPlayerDataToNearby(GUID, false)
				timestampUpdated(GUID, timestamp)
			end
		elseif lastUpdated[GUID] ~= nil then
			lastUpdated[GUID] = nil
		end
	end
end
